/* mt -- control magnetic tape drive operation
   Copyright (C) 1991, 1992, 1995, 2001, 2007, 2010 Free Software
   Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301 USA
*/

/* Modified for the Linux SCSI tape driver by Brian Mays from code
   written by Kai Makisara.
   Last Modified: Tue Apr 23 15:37:54 EDT 1996
*/

/* If -f is not given, the environment variable TAPE is used;
   if that is not set, a default device defined in sys/mtio.h is used.
   The device must be either a character special file or a remote
   tape drive with the form "[user@]system:path".
   The default count is 1.  Some operations ignore it.

   Exit status:
   0	success
   1	invalid operation or device name
   2	operation failed

   Operations (unique abbreviations are accepted):
   eof, weof	Write COUNT EOF marks at current position on tape.
   fsf		Forward space COUNT files.
		Tape is positioned on the first block of the file.
   bsf		Backward space COUNT files.
		Tape is positioned on the first block of the file.
   fsr		Forward space COUNT records.
   bsr		Backward space COUNT records.
   bsfm		Backward space COUNT file marks.
		Tape is positioned on the beginning-of-the-tape side of
		the file mark.
   asf		Absolute space to file number COUNT.
		Equivalent to rewind followed by fsf COUNT.
   eom		Space to the end of the recorded media on the tape
		(for appending files onto tapes).
   rewind	Rewind the tape.
   offline, rewoffl
		Rewind the tape and, if applicable, unload the tape.
   status	Print status information about the tape unit.
   retension	Rewind the tape, then wind it to the end of the reel,
		then rewind it again.
   erase	Erase the tape.
   fss		(SCSI tapes) Forward space COUNT setmarks.
   bss		(SCSI tapes) Backward space COUNT setmarks.
   wset		(SCSI tapes) Write COUNT setmarks at current position
		(only SCSI tape).
   eod, seod	Space to end of valid data.  Used on streamer tape
		drives to append data to the logical and of tape.
   setblk	(SCSI tapes) Set the block size of the drive to COUNT
		bytes per record.
   setdensity	(SCSI tapes) Set the tape density code to COUNT.  The
		proper codes to use with each drive should be looked
		up from the drive documentation.
   drvbuffer	(SCSI tapes) Set the tape drive buffer code to
		NUMBER.  The proper value for unbuffered operation is
		zero and "normal" buffered operation one. The meanings
		of other values can be found in the drive
		documentation or, in case of a SCSI-2 drive, from the
		SCSI-2 standard.
   stoptions	(SCSI tapes) Set the driver options bits to COUNT for
		the device.  The bits can be set by oring the
		following values: 1 to enable write buffering, 2 to
		enable asynchronous writes, 4 to enable read ahead, 8
		to enable debugging output (if it has been compiled to
		the driver).
   stwrthreshold 
		(SCSI tapes) The write threshold for the tape device
		is set to COUNT kilobytes.  The value must be smaller
		than or equal to the driver buffer size.
   seek		(SCSI tapes) Seek to the COUNT block on the tape.
		This operation is available on some Tandberg and
		Wangtek streamers and some SCSI-2 tape drives.
   tell		(SCSI tapes) Tell the current block on tape.  This
		operation is available on some Tandberg and Wangtek
		streamers and some SCSI-2 tape drives.
   densities	(SCSI tapes) Write explanation of some common density
		codes to standard output.
   datcompression
		(some SCSI-2 DAT tapes) Inquire or set the compression
		status (on/off).  If the COUNT is one the compression
		status is printed.  If the COUNT is zero, compression
		is disabled.  Otherwise, compression is enabled.

   David MacKenzie <djm@gnu.ai.mit.edu> */

#include <configmake.h>

#include <system.h>

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#ifdef HAVE_SYS_MTIO_H
# ifdef HAVE_SYS_IO_TRIOCTL_H
#  include <sys/io/trioctl.h>
# endif
# include <sys/mtio.h>
#endif
#include <sys/file.h>
#include <fcntl.h>
#include <errno.h>
#include <stdbool.h>
#include <argp.h>
#include <argp-version-etc.h>
#include <progname.h>

#ifdef HAVE_LOCALE_H
# include <locale.h>
#endif
#include <rmt-command.h>

#include <rmt.h>

#include <argmatch.h>
#include <paxlib.h>
#include "configmake.h"

#define MT_EXIT_SUCCESS 0
#define MT_EXIT_INVOP   1
#define MT_EXIT_FAILURE 2

#if defined(linux) || defined(__linux)
#define MTDATCOMP 1000         /* Random unused number.  */
#define MTDENS 1001            /* Random unused number.  */

struct densities {
  int code;
  char *name;
} density_tbl[] = {
  {0x00, "default"},
  {0x01, "NRZI (800 bpi)"},
  {0x02, "PE (1600 bpi)"},
  {0x03, "GCR (6250 bpi)"},
  {0x04, "QIC-11"},
  {0x05, "QIC-45/60 (GCR, 8000 bpi)"},
  {0x06, "PE (3200 bpi)"},
  {0x07, "IMFM (6400 bpi)"},
  {0x08, "GCR (8000 bpi)"},
  {0x09, "GCR /37871 bpi)"},
  {0x0a, "MFM (6667 bpi)"},
  {0x0b, "PE (1600 bpi)"},
  {0x0c, "GCR (12960 bpi)"},
  {0x0d, "GCR (25380 bpi)"},
  {0x0f, "QIC-120 (GCR 10000 bpi)"},
  {0x10, "QIC-150/250 (GCR 10000 bpi)"},
  {0x11, "QIC-320/525 (GCR 16000 bpi)"},
  {0x12, "QIC-1350 (RLL 51667 bpi)"},
  {0x13, "DDS (61000 bpi)"},
  {0x14, "EXB-8200 (RLL 43245 bpi)"},
  {0x15, "EXB-8500 (RLL 45434 bpi)"},
  {0x16, "MFM 10000 bpi"},
  {0x17, "MFM 42500 bpi"},
  {0x18, "TZ86"},
  {0x19, "DLT 10GB"},
  {0x1a, "DLT 20GB"},
  {0x1b, "DLT 35GB"},
  {0x1c, "QIC-385M"},
  {0x1d, "QIC-410M"},
  {0x1e, "QIC-1000C"},
  {0x1f, "QIC-2100C"},
  {0x20, "QIC-6GB"},
  {0x21, "QIC-20GB"},
  {0x22, "QIC-2GB"},
  {0x23, "QIC-875"},
  {0x24, "DDS-2"},
  {0x25, "DDS-3"},
  {0x26, "DDS-4 or QIC-4GB"},
  {0x27, "Exabyte Mammoth"},
  {0x28, "Exabyte Mammoth-2"},
  {0x29, "QIC-3080MC"},
  {0x30, "AIT-1 or MLR3"},
  {0x31, "AIT-2"},
  {0x32, "AIT-3"},
  {0x33, "SLR6"},
  {0x34, "SLR100"},
  {0x40, "DLT1 40 GB, or Ultrium"},
  {0x41, "DLT 40GB, or Ultrium2"},
  {0x42, "LTO-2"},
  {0x45, "QIC-3095-MC (TR-4)"},
  {0x47, "TR-5"},
  {0x48, "DAT160"},
  {0x80, "DLT 15GB uncomp. or Ecrix"},
  {0x81, "DLT 15GB compressed"},
  {0x82, "DLT 20GB uncompressed"},
  {0x83, "DLT 20GB compressed"},
  {0x84, "DLT 35GB uncompressed"},
  {0x85, "DLT 35GB compressed"},
  {0x86, "DLT1 40 GB uncompressed"},
  {0x87, "DLT1 40 GB compressed"},
  {0x88, "DLT 40GB uncompressed"},
  {0x89, "DLT 40GB compressed"},
  {0x8c, "EXB-8505 compressed"},
  {0x90, "SDLT110 uncompr/EXB-8205 compr"},
  {0x91, "SDLT110 compressed"},
  {0x92, "SDLT160 uncompressed"},
  {0x93, "SDLT160 comprssed"},
  {-1, NULL}};
#endif

char *cbnames[] =
{
#ifdef MT_ST_BOOLEANS
  "stoptions",
#endif
#ifdef MT_ST_WRITE_THRESHOLD
  "stwrthreshold",
#endif
  NULL
};

int count_bits[] =
{
#ifdef MT_ST_BOOLEANS
  MT_ST_BOOLEANS,
#endif
#ifdef MT_ST_WRITE_THRESHOLD
  MT_ST_WRITE_THRESHOLD,
#endif
  0
};

#ifdef MT_TAPE_INFO
  struct mt_tape_info tapes[] = MT_TAPE_INFO;
#endif

char const * const opnames[] =
{
  "eof",
  "weof",
  "fsf",
  "bsf",
  "fsr",
  "bsr",
  "rewind",
  "offline",
  "rewoffl",
  "eject",
  "status",
#ifdef MTBSFM
  "bsfm",
#endif
#ifdef MTEOM
  "eom",
  "eod",
  "seod",
#endif
#ifdef MTRETEN
  "retension",
#endif
#ifdef MTERASE
  "erase",
#endif
  "asf",
#ifdef MTFSFM
  "fsfm",
#endif
#ifdef MTSEEK
  "seek",
#endif
#ifdef MTTELL
  "tell",
#endif
#ifdef MTFSS
  "fss",
#endif
#ifdef MTBSS
  "bss",
#endif
#ifdef MTWSM
  "wset",
#endif
#ifdef MTSETBLK
  "setblk",
#endif
#ifdef MTSETDENSITY
  "setdensity",
#endif
#ifdef MTSETDRVBUFFER
  "drvbuffer",
#ifdef MT_ST_BOOLEANS
  "stoptions",
#endif
#ifdef MT_ST_WRITE_THRESHOLD
  "stwrthreshold",
#endif
#endif
#ifdef MTDATCOMP
  "datcompression",
#endif
#ifdef MTDENS
  "densities",
#endif
  NULL
};

#define MTASF 600		/* Random unused number.  */
short operations[] =
{
  MTWEOF,
  MTWEOF,
  MTFSF,
  MTBSF,
  MTFSR,
  MTBSR,
  MTREW,
  MTOFFL,
  MTOFFL,
  MTOFFL,
  MTNOP,
#ifdef MTBSFM
  MTBSFM,
#endif
#ifdef MTEOM
  MTEOM,
  MTEOM,
  MTEOM,
#endif
#ifdef MTRETEN
  MTRETEN,
#endif
#ifdef MTERASE
  MTERASE,
#endif
  MTASF,
#ifdef MTFSFM
  MTFSFM,
#endif
#ifdef MTSEEK
  MTSEEK,
#endif
#ifdef MTTELL
  MTTELL,
#endif
#ifdef MTFSS
  MTFSS,
#endif
#ifdef MTBSS
  MTBSS,
#endif
#ifdef MTWSM
  MTWSM,
#endif
#ifdef MTSETBLK
  MTSETBLK,
#endif
#ifdef MTSETDENSITY
  MTSETDENSITY,
#endif
#ifdef MTSETDRVBUFFER
  MTSETDRVBUFFER,
#ifdef MT_ST_BOOLEANS
  MTSETDRVBUFFER,
#endif
#ifdef MT_ST_WRITE_THRESHOLD
  MTSETDRVBUFFER,
#endif
#endif
#ifdef MTDATCOMP
  MTDATCOMP,
#endif
#ifdef MTDENS
  MTDENS,
#endif
};

ARGMATCH_VERIFY (opnames, operations);

const char *argp_program_bug_address = "<" PACKAGE_BUGREPORT ">";
static char doc[] = N_("control magnetic tape drive operation");
const char *program_authors[] =
  {
    "David MacKenzie",
    "Sergey Poznyakoff",
    NULL
  };

enum
  {
    RSH_COMMAND_OPTION = 256
  };

static struct argp_option options[] = {
  { "file", 'f', N_("DEVICE"), 0,
    N_("use device as the file name of the tape drive to operate on") },
  { "rsh-command", RSH_COMMAND_OPTION, N_("COMMAND"), 0,
    N_("use remote COMMAND instead of rsh") },
  { NULL }
};

char *tapedev;                   /* tape device */
char *rsh_command_option = NULL; /* rsh command */
short operation;                 /* operation code */ 
int count = 1;                   /* count */

int argcnt = 0;                  /* number of command line arguments
				    processed so far */

static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  switch (key)
    {
    case ARGP_KEY_ARG:
      switch (argcnt++)
	{
	case 0:
	  operation = XARGMATCH (N_("operation"), arg, opnames, operations);
	  break;

	case 1:
	  {
	    char *p;
	    long val = strtol (arg, &p, 0);
	    if (*p || (count = val) != val)
	      error (MT_EXIT_INVOP, 0, _("invalid count value"));
	  }
	  break;

	default:
	  argp_usage (state);
	}
      break;

    case ARGP_KEY_FINI:
      if (argcnt == 0)
	argp_usage (state);
      if (tapedev == NULL)
	{
	  tapedev = getenv ("TAPE");
	  if (tapedev == NULL)

/* Suse doesn't have /dev/tape as link to /dev/nst0 any more.
Instead it uses udev and creates different names in /dev/tape/by-id/ directory.
If it is SCSI tape storage then it creates /dev/tape/by-id/scsi--nst
If it is USB device then it creates something not predictibable:
/dev/tape/by-id/scsi-*HP_blabla*{VENDOR_SPECIFIC}*-nst
So let's use old behave which was /dev/nst0.
bnc#355241
*/
/* #ifdef DEFTAPE			* From sys/mtio.h.  * */
# define DEFSUSETAPE        "/dev/nst0"
        tapedev = DEFSUSETAPE;
	}
      break;
      
    case 'f':
    case 't':
      tapedev = arg;
      break;

    case RSH_COMMAND_OPTION:
      rsh_command_option = arg;
      break;

    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

static struct argp argp = {
  options,
  parse_opt,
  N_("operation [count]"),
  doc,
  NULL,
  NULL,
  NULL
};

void
check_type (char *dev, int desc)
{
  struct stat stats;

  if (_isrmt (desc))
    return;
  if (fstat (desc, &stats) == -1)
    stat_error (dev);
  if ((stats.st_mode & S_IFMT) != S_IFCHR)
    error (MT_EXIT_INVOP, 0, _("%s is not a character special file"), dev);
}

void
perform_operation (char *dev, int desc, short op, int count)
{
  struct mtop control;

  control.mt_op = op;
  control.mt_count = count;
  if (rmtioctl (desc, MTIOCTOP, (char*)&control) == -1)
    error (MT_EXIT_FAILURE, errno, _("%s: rmtioctl failed"), dev);
}

void
print_status (char *dev, int desc)
{
  struct mtget status;
  #ifdef MT_TAPE_INFO
    struct mt_tape_info *mt;
  #endif
 
  if (rmtioctl (desc, MTIOCGET, (char*)&status) == -1)
    error (MT_EXIT_FAILURE, errno, _("%s: rmtioctl failed"), dev);

  #ifdef MT_TAPE_INFO
    for (mt = tapes; mt->t_type; mt++)
      if (mt->t_type == status.mt_type) break;
    if (mt->t_type != 0)
      {
        printf ("drive type = %s\n", mt->t_name);
      }
    else
  #endif
 
  printf ("drive type = %d\n", (int) status.mt_type);
#if defined(hpux) || defined(__hpux)
  printf ("drive status (high) = %d\n", (int) status.mt_dsreg1);
  printf ("drive status (low) = %d\n", (int) status.mt_dsreg2);
#else
  printf ("drive status = %d\n", (int) status.mt_dsreg);
#endif
  printf ("sense key error = %d\n", (int) status.mt_erreg);
  printf ("residue count = %d\n", (int) status.mt_resid);
#if !defined(ultrix) && !defined(__ultrix__) && !defined(hpux) && !defined(__hpux) && !defined(__osf__)
  printf ("file number = %d\n", (int) status.mt_fileno);
  printf ("block number = %d\n", (int) status.mt_blkno);
#endif
#if defined(linux) || defined(__linux)
  if (status.mt_type == MT_ISSCSI1 ||
      status.mt_type == MT_ISSCSI2)
    {
      int dens, i;
      char *density;
      dens = (status.mt_dsreg & MT_ST_DENSITY_MASK) >> MT_ST_DENSITY_SHIFT;
      density = "unknown";
      for (i=0; density_tbl[i].code >= 0; i++)
       if (density_tbl[i].code == dens)
         {
           density = density_tbl[i].name;
           break;
         }
      printf("Tape block size %ld bytes. Density code 0x%x (%s).\n",
            ((status.mt_dsreg & MT_ST_BLKSIZE_MASK) >> MT_ST_BLKSIZE_SHIFT),
            dens, density);

      printf("Soft error count since last status=%ld\n",
            (status.mt_erreg & MT_ST_SOFTERR_MASK) >> MT_ST_SOFTERR_SHIFT);
      printf("General status bits on (%lx):\n", status.mt_gstat);
      if (GMT_EOF(status.mt_gstat))
       printf(" EOF");
      if (GMT_BOT(status.mt_gstat))
       printf(" BOT");
      if (GMT_EOT(status.mt_gstat))
       printf(" EOT");
      if (GMT_SM(status.mt_gstat))
       printf(" SM");
      if (GMT_EOD(status.mt_gstat))
       printf(" EOD");
      if (GMT_WR_PROT(status.mt_gstat))
       printf(" WR_PROT");
      if (GMT_ONLINE(status.mt_gstat))
       printf(" ONLINE");
      if (GMT_D_6250(status.mt_gstat))
       printf(" D_6250");
      if (GMT_D_1600(status.mt_gstat))
       printf(" D_1600");
      if (GMT_D_800(status.mt_gstat))
       printf(" D_800");
      if (GMT_DR_OPEN(status.mt_gstat))
       printf(" DR_OPEN");
      if (GMT_IM_REP_EN(status.mt_gstat))
       printf(" IM_REP_EN");
      if (status.mt_gstat != 0)
	putchar ('\n');
    }
  else
    {
      printf("gstat = %0lx\n", status.mt_gstat);
    }
#endif
}


#if defined(linux) || defined(__linux)
/*** Get and set the DAT compression (Mode Page 15) ***/

#define MODE_SENSE 0x1a
#define MODE_SELECT 0x15

int
read_mode_page(int fn, int page, int length, unsigned char *buffer,
              int do_mask)
{
  int result, *ip;
  unsigned char tmpbuffer[30], *cmd;

  memset(tmpbuffer, 0, 14);
  ip = (int *)&(tmpbuffer[0]);
  *ip = 0;
  *(ip+1) = length + 4;

  cmd = &(tmpbuffer[8]);
  cmd[0] = MODE_SENSE;
  cmd[1] = 8;
  cmd[2] = page;
  if (do_mask)
    cmd[2] |= 0x40;  /* Get changeable parameter mask */
  cmd[4] = length + 4;

  result = ioctl(fn, 1, tmpbuffer);
  if (result) {
    fprintf(stderr, "Can't read mode page. Are you sure you are root?\n");
    return 0;
  }
  memcpy(buffer, tmpbuffer + 8, length + 4);
  return 1;
}


int
write_mode_page(int fn, int page, int length, unsigned char *buffer)
{
  int result, *ip;
  unsigned char tmpbuffer[40], *cmd;

  memset(tmpbuffer, 0, 14);
  ip = (int *)&(tmpbuffer[0]);
  *ip = length + 4;
  *(ip+1) = 0;

  cmd = &(tmpbuffer[8]);
  cmd[0] = MODE_SELECT;
  cmd[1] = 0x10;
  cmd[4] = length + 4;

  memcpy(tmpbuffer + 14, buffer, length + 4);
  tmpbuffer[14] = 0;  /* reserved data length */
  tmpbuffer[18] &= 0x3f;  /* reserved bits in page code byte */

  result = ioctl(fn, 1, tmpbuffer);
  if (result) {
    fprintf(stderr, "Can't write mode page.\n");
    return 0;
  }
  return 1;
}


int
do_dat_compression(char *dev, int fn, int count)
{
  int i;
  unsigned char buffer[30], mask[30];

  if (!read_mode_page(fn, 0x0f, 16, buffer, 0)) {
    error (2, errno, "%s", dev);
  }

  if (count != -1) {
    if (count == 0)
      buffer[4+2] &= 0x7f;
    else
      buffer[4+2] |= 0x80;
/*    if (read_mode_page(fn, 0x0f, 16, mask, 1))
      for (i=3; i < 16; i++)
       buffer[4+i] |= mask[4+i];                bug #223494 */
    if (!write_mode_page(fn, 0x0f, 16, buffer)) {
      error (2, errno, "%s", dev);
    }
    if (!read_mode_page(fn, 0x0f, 16, buffer, 0)) {  /* Re-read to check */
      error (2, errno, "%s", dev);
    }
  }

  if (buffer[4+2] & 0x80)
    printf("Compression on.\n");
  else
    printf("Compression off.\n");

  return 1;
}
#endif

#ifdef MTTELL
void
print_position (dev, desc)
     char *dev;
     int desc;
{
  struct mtpos position;

  if (rmtioctl (desc, MTIOCPOS, &position) == -1)
    error (2, errno, "%s", dev);
  printf("At block %ld.\n", position.mt_blkno);

}
#endif


void
fatal_exit ()
{
  exit (MT_EXIT_INVOP);
}

int
main (int argc, char **argv)
{
  int tapedesc,i;

  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);
  
  set_program_name (argv[0]);
  argp_version_setup ("mt", program_authors);
  argmatch_die = fatal_exit;
  argp_err_exit_status = MT_EXIT_INVOP;
  if (argp_parse (&argp, argc, argv, ARGP_IN_ORDER, NULL, NULL))
    exit (MT_EXIT_INVOP);

#ifdef MTDENS
  if (operation == MTDENS)
    {
      printf("Some SCSI tape density codes:\ncode   explanation\n");
      for (i=0; density_tbl[i].code >= 0; i++)
	printf("0x%02x   %s\n", density_tbl[i].code, density_tbl[i].name);
      exit (0);
    }
#endif

  switch (operation)
    {
    case MTWEOF:
#ifdef MTERASE
    case MTERASE:
#endif
#ifdef MTWSM
    case MTWSM:
#endif
#ifdef MTSETDRVBUFFER
    case MTSETDRVBUFFER:
#endif
#ifdef MTDATCOMP
	case MTDATCOMP:
#endif
       tapedesc = rmtopen (tapedev, O_WRONLY | O_NONBLOCK, 0, rsh_command_option);
      break;

    default:
      tapedesc = rmtopen (tapedev, O_RDONLY | O_NONBLOCK, 0, rsh_command_option);
    }
  
  if (tapedesc == -1)
    error (MT_EXIT_INVOP, errno, _("%s: rmtopen failed"), tapedev);
  check_type (tapedev, tapedesc);

#ifdef MTDATCOMP
  if (operation == MTDATCOMP)
    do_dat_compression(tapedev, tapedesc, count);
  else
#endif
#ifdef MTTELL
  if (operation == MTTELL)
    print_position (tapedev, tapedesc);
  else
#endif
  {
  if (operation == MTASF)
    {
      perform_operation (tapedev, tapedesc, MTREW, 1);
      operation = MTFSF;
    }
  perform_operation (tapedev, tapedesc, operation, count);
  if (operation == MTNOP)
    print_status (tapedev, tapedesc);
  }

  if (rmtclose (tapedesc) == -1)
    error (MT_EXIT_FAILURE, errno, _("%s: rmtclose failed"), tapedev);

  exit (MT_EXIT_SUCCESS);
}

